---
celltoolbar: Diaporama
jupytext:
  notebook_metadata_filter: all
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.13
    jupytext_version: 1.15.2
kernelspec:
  display_name: Python 3 (ipykernel)
  language: python
  name: python3
language_info:
  codemirror_mode:
    name: ipython
    version: 3
  file_extension: .py
  mimetype: text/x-python
  name: python
  nbconvert_exporter: python
  pygments_lexer: ipython3
  version: 3.8.8
metadata:
  execution:
    allow_errors: true
rise:
  enable_chalkboard: true
  height: 90%
  scroll: true
  width: 90%
---

+++ {"editable": true, "slideshow": {"slide_type": ""}}

# Séance 9 : classes

+++ {"slideshow": {"slide_type": ""}, "editable": true}

## Paradigmes de programmation

- Au début du cours de L1, vous avez programmé en écrivant des instructions les unes après les autres, que l'ordinateur exécute ensuite dans l'ordre. Ce type de programmation s'appelle la **programmation impérative**.

- Plus tard dans le cours, vous avez déclaré des fonctions, qui restent fixes, et qui peuvent être appelées à différents endroits du code. Il s'agit de **programmation fonctionnelle**.

- Les deux séances qui suivent sont consacrées à un troisième paradigme de programmation, il s'agit de la **programmation orientée objet**. Nous définirons de nouveaux concepts, dont l'encapsulation, le polymorphisme et l'héritage, qui permettent de programmer différemment.

+++ {"slideshow": {"slide_type": ""}, "editable": true}

## Programmation orientée objet

- Un objet représente un ensemble symbolique contenant des données sur lui-même (**attributs**) et des fonctions (**méthodes**) permettant de manipuler ces données.

- Un objet est créé à partir d'un "moule" appelé **classe** qui dicte son comportement en définissant des attributs et des méthodes. On dit qu'un objet est une **instance** de la classe à laquelle il appartient.

- Cela permet de s'abstraire du fonctionnement interne de l'objet. On peut agir sur les attributs d'un objet en passant par ses méthodes, même si on ne connait pas le détail de son implémentation. Cela permet donc de séparer le travail sur des objets différents sans risque de conflit.

+++ {"slideshow": {"slide_type": ""}, "editable": true}

## Exemple abstrait

On définit une classe Voiture, qui contient :

- des attributs : couleur, marque, quantité d'essence, position, ...

- des méthodes : conduire, faire le plein, ...

On peut ensuite créer des instances de cette classe :

- Nom : Clio

- Attributs : couleur=rouge, marque=Renault, Quantité d'essence=20L, Position=(48.7, 2.1)

- Méthodes : conduire, faire le plein, ...

Utiliser une méthode peut modifier les attributs d'un objet : quand vous utilisez la méthode conduire, vous pourrez par exemple modifer la position, tout en faisant baisser la quantité d'essence.

+++ {"slideshow": {"slide_type": ""}, "editable": true}

## En Python

Tout en Python est un objet ! Quand on interroge le type d'un objet, on obtient une information sur sa classe d'appartenance. Par exemple une chaîne de caractères est une instance de la classe `str`:

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
editable: true
slideshow:
  slide_type: ''
---
a="BLA"
print(type(a))
```

+++ {"editable": true, "slideshow": {"slide_type": ""}}

`numpy` est une instance de la classe `module` :

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
editable: true
slideshow:
  slide_type: ''
---
import numpy
print(type(numpy))
```

+++ {"editable": true, "slideshow": {"slide_type": ""}}

On peut aussi vérifier si un objet est issu d'une certaine classe :

```{code-cell} ipython3
---
editable: true
slideshow:
  slide_type: ''
---
isinstance(a,str)
```

+++ {"slideshow": {"slide_type": "slide"}}

## Méthodes

Regardons l'objet `a="blanc bleu rouge"`. Cet objet possède ses propres
data ("blanc bleu rouge") mais possède aussi des méthodes pour les
manipuler :

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
a.upper()
```

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
a.split()
```

+++ {"slideshow": {"slide_type": "slide"}}

Des méthodes peuvent modifier l'objet de départ:

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
a=[3, 9, 10, 12, 45]
a.append(5)
print(a)
```

ou non :

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
a="bla"
a.upper()
```

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
print(a)
```

+++ {"slideshow": {"slide_type": "slide"}}

## Construction des classes

On peut construire des classes qui définissent des objets adaptés aux
problèmes auxquels on s'intéresse et qui permettent d'aller au delà des
types de base.

Un exemple minimal :

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
class test_class:
    pass
```

nous avons généré une classe `test_class` et nous pouvons maintenant
l'utiliser pour construire des instances de la classe:

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
instance1 = test_class()
print(type(instance1))
```

+++ {"slideshow": {"slide_type": "slide"}}

## Attributs

On peut définir ensuite des attributs de classe:

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
class test_class:
    attribute1 = "Ceci est un attribut"
```

qui seront identiues pour chaque instance de la classe et appelés sous
la forme `instance.attribut`:

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
instance1 = test_class()
instance1.attribute1
```

+++ {"slideshow": {"slide_type": "slide"}}

## Méthodes

Une fonction définie à l'intérieur d'une classe s'appelle méthode et
elle est accessible par toute instance de la classe.

Exemple:

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
class test_class:
    def test_method(self):
        print("Méthode test de ma classe test")
```

cette classe définit une méthode `test_method` qui prend un seul argument, l'instance même (`self`).

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
a=test_class()
a.test_method()
```

+++ {"slideshow": {"slide_type": "slide"}}

## Le constructeur

La méthode `__init__()` permet de construire des objets à partir de la classe. Elle est appelée chaque fois qu'un nouvel objet d'une classe est créé (c'est à dire à l'**instanciation**), ce qui permet de passer des attributs à la création de l'objet.

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
class test:
    def __init__(self,inputdata):
        self.data=inputdata
```

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
a=test(3)
print(a.data)
```

`self` est une référence à l'instance en cours. `self.data=` crée donc un attribut d'instance, c'est à dire un attribut qui sera différent dans différents objets. Une fois un objet créé, on peut accéder aux attributs d'instances dans toutes les méthodes, sans qu'ils y soient définis.

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
class test:
    def __init__(self,inputdata):
        self.data=inputdata
```

dans cet exemple vous êtes obligés de rentrer une valeur à la création
de votre instance car son initialisation requière la génération de
l'attribut `data`. Cependant, comme pour toute fonction, vous pouvez
définir des valeur par défaut :

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
class test:
    def __init__(self,inputdata=0):
        self.data=inputdata
```

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
a=test()
print(a.data)
```

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
a=test(3)
print(a.data)
```

+++ {"slideshow": {"slide_type": "slide"}}

## Exemple 1

On construit une classe qui

-   à la création d'une instance prend une chaîne de caractères et
    construit un attribut `data` qui est une liste des mots de la
    chaîne;

-   définit une méthode `add` qui permet d'ajouter d'autres mots à cette
    liste.

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
class test:
    def __init__(self,inputstring):
        self.data=inputstring.split()       
    def add(self,inputstring):
        self.data += inputstring.split()
```

+++ {"slideshow": {"slide_type": "slide"}}

Utilisation :

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
a=test("Atchoum, Dormeur, Grincheux, Joyeux")
print(a.data)
```

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
a.add("Prof, Simplet, Timide")
print(a.data)
```

+++ {"slideshow": {"slide_type": "slide"}}

## Exemple 2

On définir une classe `BankAccount` qui

-   créé des instances dans lesquelles enregistrer le nom du client, le
    numéro de compte et le solde;

-   définit une méthode `deposit` pour déposer des fonds;

-   définit une méthode `withdraw` pour retirer des fonds.

+++ {"slideshow": {"slide_type": "slide"}}

Définition de la classe `BankAccount` :

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
class BankAccount:
    def __init__(self,customer,account_number,balance=0):
        self.customer=customer
        self.account_number=account_number
        self.balance=balance
    def deposit(self, amount):
        if amount >0 :
             self.balance += amount
        else:
             print("invalid deposit amount")
    def withdraw(self, amount):
        if amount >0 :
             self.balance -= amount
        else:
             print("invalid withdraw amount")
```

+++ {"slideshow": {"slide_type": "slide"}}

Utilisation de la classe `BankAccount` :

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
account1=BankAccount("Uncle Scrooge", "101010", 1)
account1.customer
```

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
account1.balance
```

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
account1.deposit(1E15)
account1.balance
```

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
account1.withdraw(2)
account1.balance
```

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
account1.withdraw(-2)
```

## Encapsulation

L'encapsulation consiste à regrouper des variables avec des routines qui agissent sur ces variables ou qui les utilisent. Cela permet de contrôler la façon dont on accède à ces variables, et de former des sortes de boîtes noires. C'est ce concept qui motive la création des attributs et des méthodes.

Cette logique est poussée très loin dans d'autres langages de programmation orientés objets, comme C++ et Java. Dans ces langages, on peut protéger les attributs pour qu'il soit impossible d'y accéder en dehors de la classe. On implémente des fonctions pour renvoyer la valeur des attributs (**getter**) ou pour la modifier (**setter**), ce qui permet de contrôler ce qui y est mis (par exemple pour éviter qu'un nombre soit négatif). Mais cela multiplie le nombre de méthodes à implémenter.

En Python, on privilégie la lisibilité du code. Pour éviter la multiplication des getters et setters, les attributs ne sont pas protégés. Il est même possible de créer un attribut en dehors d'une classe, mais cela reste déconseillé.

Il existe tout de même des façons de protéger les attributs en Python. Les attributs dont le nom commence par un ou deux underscores sont **non-publics**, c'est à dire qu'il faut par convention éviter de les utiliser en dehors de la classe. De plus, si l'on essaye d'accéder à un attribut dont le nom commence par deux underscores en dehors de sa classe, on obtient une erreur. Cette erreur est due au fait que l'attribut a un nom différent à l'intérieur et à l'extérieur de la classe, ce qui s'appelle **name mangling**.

Pour aller plus loin, on peut utiliser le décorateur property (hors programme).

+++

## Espaces de noms

Un autre aspect de l'encapsulation en Python est qu'une classe crée son propre espace de noms (namespace), ce qui permet d'encapsuler les attributs et méthodes sans risque de conflit avec le reste du code. Un espace de noms est ce qui lie les noms aux objets, que ce soit des variables ou des fonctions :

```{code-cell} ipython3
class test:
    def __init__(self,inputdata):
        self.data=inputdata
        
    def affiche(self):
        print("Affichage dans la classe")

def affiche():
    print("Affichage hors de la classe")
```

```{code-cell} ipython3
data = "Atchoum"
a = test("Grincheux")
print(data)
print(a.data)
```

```{code-cell} ipython3
a.affiche()
affiche()
```

De la même façon, une fonction définie un espace de noms, et il est possible de définir une variable **locale** à cette fonction.

+++

## Composition

Les attributs sont des objets, on peut donc y mettre des instances des classes qu'on a créées. Dans notre classe `BankAccount`, on peut par exemple remplacer le nom du client par un objet client :

```{code-cell} ipython3
class Customer:
    def __init__(self,name,premium=False):
        self.name=name
        self.premium=premium

class BankAccount:
    def __init__(self,customer,account_number,balance=0):
        self.customer=customer
        self.account_number=account_number
        self.balance=balance
```

```{code-cell} ipython3
scrooge = Customer("Uncle Scrooge", premium=True)
account1=BankAccount(scrooge, "101010", 1)
account1.customer
```

## Documenter les classes

Du texte écrit entre triples guillemets après la définition de la classe
et des méthodes constitue une documentation. Celle ci peut être appelée
grâce à la fonction help.

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
class BankAccount:
    """ A class representing a bank customer """
    def __init__(self,balance=0):
        """ Initialize the Bank account class """
        self.balance=balance
    def deposit(self, amount):
        """ Deposit a given amount """
        if amount >0 :
            self.balance += amount
    def withdraw(self, amount):
        """ Withdraw a given amount """
        if amount >0 :
            self.balance -= amount
```

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
help(BankAccount)
```

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
help(BankAccount.deposit)
```

```{code-cell} ipython3
---
codeCellConfig:
  lineNumbers: true
slideshow:
  slide_type: '-'
tags: [raises-exception]
---
help(BankAccount.withdraw)
```

Cette documentation peut aussi être automatiquement transformée en site web de documentation. C'est comme cela que sont produits certains des sites web que vous consultez, comme la documentation de `numpy`.

+++ {"editable": true, "slideshow": {"slide_type": ""}}

## Création de modules

Une fois que vous avez créé une classe (ou un ensemble de fonctions), vous pouvez avoir envie de la réutiliser dans plusieurs notebooks. Pour cela, il est possible de créer un **module**, comme `math` ou `numpy`, en écrivant votre classe dans un ficher .py dans le même répertoire que votre notebook. Il suffit ensuite de l'importer comme d'habitude :

```{code-cell} ipython3
---
editable: true
slideshow:
  slide_type: ''
---
import monModule
```

+++ {"editable": true, "slideshow": {"slide_type": ""}}

Bien documenter les classes et les fonctions de son module est crucial pour ne pas avoir à constamment aller consulter son code.
