---
jupytext:
  formats: ipynb,md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.13
    jupytext_version: 1.15.2
kernelspec:
  display_name: Python 3 (ipykernel)
  language: python
  name: python3
---

+++ {"nbgrader": {"grade": false, "grade_id": "cell-d76bcf6bac33365a", "locked": true, "schema_version": 3, "solution": false, "task": false}}

# TD9 : introduction aux classes

+++ {"nbgrader": {"grade": false, "grade_id": "cell-17331d0f6fa8cc60", "locked": true, "schema_version": 3, "solution": false, "task": false}}

## Introduction

Le but de cet exercice est de vous familiariser avec la terminologie des classes et leur utilisation. Regarder l’exemple de la classe ``atom`` suivante qui permet de construire des objets de type atome:

```{code-cell} ipython3
# Petit rappel de PANDA
import numpy as nb
import pandas as pd

elements=pd.read_table("data/elements.txt")
print(elements[0:20])
print("\n")

Z = 35
# on extrait du dataframe "elements" la ligne correspondant à Z=35 (attention, le sdonées sont au format string)
line = elements[elements["Z"] == str(Z)]
print(line)
print("\n")
# on extrait de la ligne le Nom et le Symbole de l'élément
print("Nom : ",line["Name"].values[0], "  Symbole: ",line["Symbol"].values[0])
```

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-e94d06732cdb4b21
  locked: true
  schema_version: 3
  solution: false
  task: false
---
import numpy as nb
import pandas as pd

class atom:
    elements=pd.read_table("data/elements.txt")
    
    def __init__(self,Z):
        self.Z=Z
        self.line = self.elements[self.elements["Z"] == str(self.Z)] # extrait de "elements" la ligne correspondant à Z
        
    def name(self):
        return self.line["Name"].values[0]
        
    def symbol(self):
        return self.line["Symbol"].values[0]
```

+++ {"nbgrader": {"grade": false, "grade_id": "cell-4daa4b2ae2a046c2", "locked": true, "schema_version": 3, "solution": false, "task": false}}

Construisez une instance d'un atom avec Z=35 et A=81.

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-c4694d946334ffe8
  locked: false
  schema_version: 3
  solution: true
  task: false
---
### BEGIN SOLUTION
atom1=atom(35)
### END SOLUTION
```

+++ {"nbgrader": {"grade": false, "grade_id": "cell-677519cc5d625f8f", "locked": true, "schema_version": 3, "solution": false, "task": false}}

Afficher les attributs Z et A de cet atome.

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-6afbd3a86900ec10
  locked: false
  schema_version: 3
  solution: true
  task: false
---
### BEGIN SOLUTION
print("Z = ",atom1.Z)
### END SOLUTION
```

+++ {"nbgrader": {"grade": false, "grade_id": "cell-d7819f12276865c9", "locked": true, "schema_version": 3, "solution": false, "task": false}}

Grâce aux méthodes respectives, afficher le nom et le symbole de cette instance.

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-ec7867c4f39ec274
  locked: false
  schema_version: 3
  solution: true
  task: false
---
### BEGIN SOLUTION
print("Nom: ", atom1.name(),"   Symbole: ",atom1.symbol())
### END SOLUTION
```

+++ {"nbgrader": {"grade": false, "grade_id": "cell-486ab53d808cd390", "locked": true, "schema_version": 3, "solution": false, "task": false}}

## Exercise 1: évaluer un QCM

On veut définir une classe qui permet de vérifier un questionnaire à choix multiples. Le questionnaire aura N questions avec 3 possibles réponses: 1,2,3. Une question sans réponse sera indiquée par un 0.
1. Définissez une classe <tt>quiz</tt> qui, à la création d'une instance, initialise un attribut <tt>responses</tt> qui sera un questionnaire vierge. Il s'agira d'une liste de N fois la valeur 0.  
2. Définissez une méthode <tt>Q</tt> qui prends comme arguments le numéro de la question et la reponse donnée. Cette fonction vérifie si la valeur est valable (un numéro compris entre 0 et 3). Si la condition est satisfaite, la valeur est enregistré à la position correspondante de l'attribut <tt>responses</tt>, autrement le logiciel affiche la phrase "wrong input". 
3. On voudra noter le questionnaire en respectant les conditions suivantes:
    * réponse correcte +2 points 
    * réponse fausse -1 point 
    * pas de réponse 0 points 
    
   Pour ceci vous devez définir une nouvelle méthode <tt>score</tt> qui compare les valeurs enregistrées dans l'attribut <tt>responses</tt> avec ceux d'une liste de référence. Pour cet exercice vous pouvez utiliser la liste des bonnes réponses définies dans la cellule suivante.

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-c149409fc94451fd
  locked: true
  schema_version: 3
  solution: false
  task: false
---
bonnes_reponses = [1,3,2,3,2,1,3,2,2,1]
```

+++ {"nbgrader": {"grade": false, "grade_id": "cell-6297b3530caada47", "locked": true, "schema_version": 3, "solution": false, "task": false}}

Definissez votre classe dans la cellule suivante:

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-1b3d746fbc6f012b
  locked: false
  schema_version: 3
  solution: true
  task: false
---
### BEGIN SOLUTION

class quiz:
    def __init__(self):
        self.responses = len(bonnes_reponses)*[0]
    def Q(self,question_number, value):
        if value not in range(4):
            print("wrong input")
        else:
            self.responses[question_number] = value        
    def score(self):
        points=0
        for i in range(len(bonnes_reponses)):
            if self.responses[i]==0:
                points+=0
            elif self.responses[i]==bonnes_reponses[i]:
                points+=2
            else:
                points-=1
        return points
### END SOLUTION
```

+++ {"nbgrader": {"grade": false, "grade_id": "cell-5463471478cb90e9", "locked": true, "schema_version": 3, "solution": false, "task": false}}

Marie à répondu au questionnaire de la façon suivante:
0,3,0,1,2,3,0,0,0,0. En exécutant la cellule suivante vous pouvez vérifier si vous avez trouvé la bonne note.

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-9dfc547f44501595
  locked: true
  schema_version: 3
  solution: false
  task: false
---
marie=quiz()
marie.Q(1,3)
marie.Q(3,1)
marie.Q(4,2)
marie.Q(5,3)
marie.score()==2
```

+++ {"nbgrader": {"grade": false, "grade_id": "cell-0b5ed26b79f330ef", "locked": true, "schema_version": 3, "solution": false, "task": false}}

## Exercice 2 : centre de masse

Implémentez une classe ``body`` qui construit un objet avec un attribut ``M`` pour indiquer sa masse et un attribut ``position`` pour indiquer sa position dans un espace bidimensionnel. L’attribut masse doit être spécifié à la génération de l’objet, l’attribut position prends la valeur (0,0) si aucune valeur est rentrée.

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-08c60037645b6b6f
  locked: false
  schema_version: 3
  solution: true
  task: false
---
### BEGIN SOLUTION

import numpy as np

class body:
    def __init__(self,masse,position=np.zeros(2)):
        self.M=masse
        self.position=position
### END SOLUTION
```

+++ {"nbgrader": {"grade": false, "grade_id": "cell-5168b774eb4041b8", "locked": true, "schema_version": 3, "solution": false, "task": false}}

* Construisez maintenant une classe ``ensemble`` pour collecter des objets ``body``.  Une entité est générée avec un attribut ``body_list``  qui corresponds à une liste vite. 
* Définissez une méthode ``add`` qui vous permet d’ajouter un objet body à votre ``body_list``. La méthode contrôle si l’objet appartiens à la classe body, si c’est vrai elle ajoute l’objet à ``body_list``, autrement elle vous affiche une message d’erreur. 
* Définissez une méthode ``remove`` qui vous permet de retirer un élément de ``body_list`` en donnant sa position dans la liste.
* Définissez une méthode ``max`` et une méthode ``min`` qui vous retournent respectivement les valeur de la plus grosses et de la plus petite masse dans la liste. Remarque : il faudra créer une liste à partir des attributs `M` des objets de votre `body_list`, et chercher la valeur des extrêmes dans cette nouvelle liste.
* Définissez une méthode ``plot`` qui vous permet d’afficher la position des objets sur le plan.  Pour faire ceci vous pouvez pour chaque objet utiliser la fonction scatter et choisissez une seule couleur pour les points. 
* (Facultatif) Il est souhaitable de pouvoir représenter la valeur de la masse des objets par la taille des points  grâce à l’option s=valeur de la fonction scatter. Vous pouvez utiliser pour les tailles  l’expression suivante:
$$size =A\sqrt{\frac{M-M_{min}}{M_{max}-M_{min}}}+B$$
où $A$ et $B$ sont des entiers choisi pour avoir une taille des points raisonable, par example $A=1000$ et $B=1$.
* Définissez une méthode ``CM`` qui vous donne la position du centre de masse du système:
$$ R= \frac{1}{M_{tot}}\sum_{i=1}^n m_i \mathbf{r}_i$$
* Modifier maintenant votre méthode plot pour que le graphe affiche en plus des positions des masses, la position du centre de masse représentée par un point rouge.

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-d6b219d1028261a6
  locked: false
  schema_version: 3
  solution: true
  task: false
---
### BEGIN SOLUTION

import matplotlib.pyplot as plt

class ensemble: 
    def __init__(self):
        self.body_list=[]
    def add(self,input_body):
        if type(input_body)==body:
            self.body_list.append(input_body)
        else:
            print("This is not a body object")
    def remove_index(self,index):
        ''' enleve de la liste l objet à la position index '''
        self.body_list.pop(index)
    def remove_objet(self, objet):
        ''' enlève de la liste l objet passé en argument '''
        self.body_list.remove(objet)
    def max(self):
        l=[x.M for x in self.body_list]
        return max(l)
    def min(self):
        l=[x.M for x in self.body_list]
        return min(l)
    def plot(self):
        for x in self.body_list:
            size=np.sqrt((x.M-self.min())/(self.max()-self.min()))*1000+1
            plt.scatter(x.position[0],x.position[1],s=size,c='b')
        plt.scatter(self.CM()[0],self.CM()[1],c="r")
        plt.show()
    def CM(self):
        Mtot = 0
        R = np.zeros(2)
        for x in self.body_list:
            Mtot += x.M
            R += x.M*x.position
        return R/Mtot
### END SOLUTION
```

+++ {"nbgrader": {"grade": false, "grade_id": "cell-0924dd00bd7d6f45", "locked": true, "schema_version": 3, "solution": false, "task": false}}

Regardez maintenant l’exemple suivant dans lequel on construit un système de masses en utilisant la classe que vous avez défini :

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-4a3a2d96f4bde64f
  locked: true
  schema_version: 3
  solution: false
  task: false
---
masse1=body(10,np.array([1,2]))
masse2=body(30,np.array([2,5]))
masse3=body(5,np.array([2,3]))
masse4=body(20,np.array([2,2]))
system=ensemble()
system.add(masse1)
system.add(masse2)
system.add(masse3)
system.plot()
system.add(masse4)
system.remove_index(1)
system.remove_objet(masse3)
system.plot()
```

+++ {"nbgrader": {"grade": false, "grade_id": "cell-c2db6d12272f176f", "locked": true, "schema_version": 3, "solution": false, "task": false}}

Vous pouvez facilement opérer sur votre entité ou directement sur les attributs des objets auxquels votre entité se réfère :

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-bfe8d566df44fdc4
  locked: true
  schema_version: 3
  solution: false
  task: false
---
system.add(body(4))
system.plot()
```

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-32f84246f85f9464
  locked: true
  schema_version: 3
  solution: false
  task: false
---
masse1.M=50
system.plot()
```

+++ {"nbgrader": {"grade": false, "grade_id": "cell-646e5707f4e560db", "locked": true, "schema_version": 3, "solution": false, "task": false}}

## Exercice 3 : polymères

Un modèle simple de polymère peut être construit en considérant une séquence de segments orientés de manière aléatoire (modèle dit de "vol aléatoire").

Dans ce problème on définira une classe `Polymer` pour décrire un tel polymère, dans laquelle un ensemble de points sont tirés aléatoirement sur une sphère et stockés dans un tableau. Par exemple, soient  $P1,P2,P3,P4...$ ces points. Le premier segment vaut $P1P2$, le second $P2P3$, etc. Les coordonnées des points $P1,P2,P3,P4...$ de la sphère sont dans un attribut de la classe.
* Une instance de la classe `Polymer` sera initialisée avec les valeurs du nombre de segments `N` et du rayon de la sphère `a`. La méthode d'initialisation appelle une méthode `make_polymer` pour construire l'ensemble des positions des extrémités des segments. 
* Dans ce problème le polymère est décrit dans un espace tridimensionnel et vous devriez donc utiliser des coordonées spheriques et deux angles aléatoires $\theta$ et $\phi$.
* Vous pouvez implémenter une méthode ``plot`` que vous permet de visualiser à trois dimensions le polymère généré. Pour ceci vous devez d’abord importer le module ``mplot3d``:

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-0f5771c60a8eb06c
  locked: true
  schema_version: 3
  solution: false
  task: false
---
from mpl_toolkits import mplot3d
```

+++ {"nbgrader": {"grade": false, "grade_id": "cell-f1bc44bb8666c9c8", "locked": true, "schema_version": 3, "solution": false, "task": false}}

ensuite à l’intérieur de votre méthode ``plot`` vous pouvez utiliser les fonctions suivantes :
```
ax = plt.axes(projection='3d')
ax.plot3D(x,y,z)
```

+++ {"nbgrader": {"grade": false, "grade_id": "cell-07ef2702d0ea91ba", "locked": true, "schema_version": 3, "solution": false, "task": false}}

où x, y et z représentent les listes des coordonnée des extrêmes des segments de votre polymère.

+++ {"nbgrader": {"grade": false, "grade_id": "cell-0686b8da5c3272f4", "locked": true, "schema_version": 3, "solution": false, "task": false}}

* La classe implémentera aussi une méthode `R` qui calculera la distance de bout en bout de la chaîne polymérique (somme des longueurs de tous les segments). On pourra utiliser la fonction _linalg.norm_ de la libraire _numpy_

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-597f12e374516c29
  locked: false
  schema_version: 3
  solution: true
  task: false
---
### BEGIN SOLUTION

import numpy as np
import matplotlib.pyplot as plt

class Polymer:
    def __init__(self,N,a):
        self.N=N
        self.a=a
        self.position=[]
        self.make_polymer()
    def make_polymer(self):
        for i in range(self.N):
            theta=np.random.uniform(0,np.pi)
            phi=np.random.uniform(0,2*np.pi)
            x= self.a * np.sin(theta) * np.cos(phi)
            y= self.a * np.sin(theta) * np.sin(phi)
            z= self.a * np.cos(theta)
            r=np.array([x,y,z])
            self.position.append(r)
    def plot(self):
        ax = plt.axes(projection='3d')
        x=[x[0] for x in self.position]
        y=[x[1] for x in self.position]
        z=[x[2] for x in self.position]
        ax.plot3D(x,y,z)
    def R(self):
        ''' Longueur totale de la molécule (comme sur les segments)'''
        return np.linalg.norm(self.position)
    
a=Polymer(25,3)
a.plot()
### END SOLUTION
```

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-01f690ebc196ae92
  locked: true
  schema_version: 3
  solution: false
  task: false
---
print(a.R())
```

+++ {"nbgrader": {"grade": false, "grade_id": "cell-f1b898e7327840c9", "locked": true, "schema_version": 3, "solution": false, "task": false}}

## Exercice 4 : classe pour l'analyse de données
On se propose ici de créer une classe Python qui permettra de faciliter l'analyse des données contenues dans un fichier.

Ce genre de classes est notamment utilisé dans l'analyse de résultats expérimentaux dans les laboratoires de physique.

> Note : la classe que vous allez créer sera modifiée au fur et à mesure de l'exercice. Il faudra donc retourner à cette cellule plusieurs fois.

> Note : cette classe sera réutilisée dans la séance suivante.

```{code-cell} ipython3
import numpy as np, matplotlib.pyplot as plt
from scipy.optimize import curve_fit
```

1. Créez une nouvelle classe `Donnees` :
   * À l'initialisation, cette classe prend un argument optionnel `chemin_fichier` (de type `str`) qui représente le chemin d'un fichier de données, de valeur par défaut `""`
   * Si la longueur de `chemin_fichier` est strictement supérieure à zéro, on utilise ensuite la fonction `np.loadtxt` pour importer les données depuis le fichier.
   * La 1ère colonne sera affectée à l'attribut `t` de l'instance, la 2ème colonne sera affectée à l'attribut `y` de l'instance.

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-fd66b56c2b6b3e1b
  locked: false
  schema_version: 3
  solution: true
  task: false
---
### BEGIN SOLUTION
def modele(t, a, b):
    return a*t + b

class Donnees():
    def __init__(self, chemin_fichier=""):
        if len(chemin_fichier)>0:
            self.t, self.y = np.loadtxt(chemin_fichier)
    
    def plot(self, ajustement=True, **kwargs):
        fig, ax = plt.subplots(1, 1)
        ax.plot(self.t, self.y, **kwargs)
        if ajustement:
            a, b, sigma_a, sigma_b = self.ajuster()
            t_plot = np.linspace(self.t[0], self.t[-1])
            y_plot = modele(t_plot, a, b)
            ax.plot(t_plot, y_plot,
                    label="Ajustement (a={:.2e}±{:.2e}, b={:.2e}±{:.2e})".format(a, sigma_a, b, sigma_b))
        ax.set_xlabel("t [s]")
        ax.set_ylabel("y")
        ax.grid(True)
        ax.legend(loc="best")
    
    def filtrer(self, lims):
        filtre = (self.t >= lims["t"][0]) & (self.t <= lims["t"][1]) & (self.y >= lims["y"][0]) & (self.y <= lims["y"][1])
        nouv_donnees = self.__class__()
        nouv_donnees.t = self.t[filtre]
        nouv_donnees.y = self.y[filtre]
        return nouv_donnees

    def moyenne(self, afficher=False):
        m, e = np.mean(self.y), np.std(self.y)
        if afficher: print("Moyenne : {:.2f} ± {:.2f}".format(m, e))
        return m, e
    
    def ajuster(self):
        t, y = self.t, self.y
        a_guess = (y[-1] - y[0]) / (t[-1] - t[0])
        b_guess = self.moyenne(afficher=False)[0]
        popt, pcov = curve_fit(modele, t, y, p0=(a_guess, b_guess))
        return *popt, np.sqrt(pcov[0,0]), np.sqrt(pcov[1,1])
### END SOLUTION
```

2. Définissez une méthode `plot` pour votre classe :
   * Cette méthode prendra un nombre indéfini d'arguments optionnels, ce que vous pouvez réaliser en écrivant `**kwargs` comme dernier argument de la méthode.
   * La méthode `plot` créera un graphe dans lequel on représentera `self.y` en fonction de `self.t`, puis on rajoutera des noms aux axes, ainsi qu'une grille.
   * Les éventuels arguments optionnels, `**kwargs`, seront passés à la fonction `plot` de `matplotlib`, comme dernier argument.

+++

3. Utilisez votre classe pour importer et représenter graphiquement les données en "dents de scie" contenues dans le fichier `"data/dents_de_scie.csv"`.

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-458dacb5b3c47461
  locked: false
  schema_version: 3
  solution: true
  task: false
---
### BEGIN SOLUTION
fichier = r"data/dents_de_scie.csv"
d = Donnees(chemin_fichier=fichier)
d.plot(ajustement=False)
### END SOLUTION
```

4. Définissez une méthode `moyenne` :
   * Cette méthode prend un argument optionnel `afficher` (de type `bool`).
   * On calcule la moyenne et l'écart-type de `self.y`.
   * Si `afficher` est `True` alors on affiche la valeur de la moyenne et de l'écart-type (`print`).
   * Enfin, on renvoie la valeur de la moyenne et de l'écart-type.

+++

5. Définissez une méthode `filtrer` :
   * Cette méthode prend comme argument un dictionnaire `lims`, avec deux clés `"t"` et `"y"` dont les valeurs sont des *tuples*.
   * Cette méthode va "filtrer" les données : créez deux nouveaux tableaux contenant uniquement les valeurs de `self.t` et `self.y` comprises dans les intervalles définis par les *tuples* dans `lims` (utilisez par exemple le *slicing* de `numpy` avec une condition, voir séance précédente).
   * Cette méthode créera ensuite une nouvelle instance de la classe `Donnees`, pour laquelle on ne renseignera pas l'argument `chemin_fichier`.
   * Il faudra ensuite modifier les attributs `t` et `y` de cette nouvelle instance pour qu'ils correspondent aux données filtrées.
   * Enfin, la méthode renvoie la nouvelle instance ainsi créée.

+++

6. Utilisez votre méthode `filtrer` pour créer une nouvelle instance de `Donnees` sur laquelle vous n'aurez plus que les points d'une seule des rampes des données en dents de scie.

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-42f1900569746bf0
  locked: false
  schema_version: 3
  solution: true
  task: false
---
### BEGIN SOLUTION
e = d.filtrer(lims={"t": (-5.5, -2.5), "y": (-np.inf, np.inf)})
e.plot(ajustement=False)
### END SOLUTION
```

7. Définissez une méthode `ajuster` :
   * Utilisez la fonction `curve_fit` du module `scipy.optimize` pour ajuster les données par une droite affine. Les paramètres initiaux peuvent être estimés avec une méthode que vous expliciterez (utilisez par exemple la méthode `moyenne`).
   * La méthode renvoie la liste des valeurs optimales trouvées par l'algorithme pour `a` et `b`, ainsi que les incertitudes trouvées par l'algorithme sur ces deux valeurs.

+++

8. Modifiez votre méthode `plot` en lui ajoutant un argument optionnel `ajustement` (de type `bool`) : s'il vaut `True`, alors la méthode `plot` superposera au graphe des données l'ajustement affine réalisé avec la méthode `ajuster`.

+++

9. Utilisez la méthode `plot` ainsi modifiée pour représenter les données et l'ajustement, et affichez sur le graphe les coefficients `a` et `b` extraits de l'ajustement.

```{code-cell} ipython3
---
nbgrader:
  grade: false
  grade_id: cell-1cfe0423fa1c4fea
  locked: false
  schema_version: 3
  solution: true
  task: false
---
### BEGIN SOLUTION
e.plot(ajustement=True)
### END SOLUTION
```
